# Generated by Django 4.2.20 on 2025-04-16 13:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Platform",
            fields=[
                ("id", models.UUIDField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=400)),
                ("abbrev", models.CharField(blank=True, max_length=50)),
                (
                    "content_provider_name",
                    models.CharField(blank=True, max_length=400, null=True),
                ),
                ("website", models.URLField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="Report",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "counter_release",
                    models.PositiveSmallIntegerField(
                        choices=[(4, "C4"), (5, "C5"), (51, "C51")]
                    ),
                ),
                ("report_id", models.CharField(max_length=20)),
            ],
            options={"unique_together": {("counter_release", "report_id")}},
        ),
        migrations.CreateModel(
            name="SushiService",
            fields=[
                ("id", models.UUIDField(primary_key=True, serialize=False)),
                (
                    "counter_release",
                    models.PositiveSmallIntegerField(
                        choices=[(4, "C4"), (5, "C5"), (51, "C51")]
                    ),
                ),
                ("url", models.URLField(blank=True, null=True)),
                (
                    "ip_address_authorization",
                    models.BooleanField(
                        blank=True,
                        help_text="Access restricted based on IP address",
                        null=True,
                    ),
                ),
                (
                    "api_key_required",
                    models.BooleanField(
                        blank=True, help_text="Is api key required", null=True
                    ),
                ),
                (
                    "platform_attr_required",
                    models.BooleanField(
                        blank=True, help_text="Is platform attr required", null=True
                    ),
                ),
                (
                    "requestor_id_required",
                    models.BooleanField(
                        blank=True, help_text="Is requestor_id required", null=True
                    ),
                ),
                (
                    "platform",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="sushi_services",
                        to="celus_registry.platform",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ReportToPlatform",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "platform",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="celus_registry.platform",
                    ),
                ),
                (
                    "report",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="celus_registry.report",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="platform",
            name="reports",
            field=models.ManyToManyField(
                related_name="platforms",
                through="celus_registry.ReportToPlatform",
                to="celus_registry.report",
            ),
        ),
    ]



from django.contrib import admin
from django.utils.html import format_html

from .models import Notification, Platform, SushiService


class ReportsInline(admin.TabularInline):
    model = Platform.reports.through

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False


class SushiServiceInline(admin.TabularInline):
    show_change_link = True
    model = SushiService

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False


@admin.register(Platform)
class PlatformAdmin(admin.ModelAdmin):
    search_fields = ("name", "abbrev", "id")
    inlines = (SushiServiceInline, ReportsInline)
    list_display = ("id", "abbrev", "name", "registry")
    readonly_fields = ("id", "abbrev", "name")

    def registry(self, obj: Platform):
        return format_html(f'<a href="{obj.registry_url}">Registry</a>')

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False


@admin.register(SushiService)
class SushiServiceAdmin(admin.ModelAdmin):
    search_fields = ("url", "id")
    list_display = (
        "id",
        "url",
        "ip_address_authorization",
        "api_key_required",
        "platform_attr_required",
        "requestor_id_required",
    )

    readonly_fields = ("id", "url", "counter_release", "platform")

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def get_actions(self, *args, **kwargs):
        return []


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    search_fields = ("subject", "message")
    list_display = (
        "id",
        "type",
        "source",
        "start_date",
        "end_date",
        "published_date",
        "last_modified",
        "sushi_service",
        "subject",
    )
    readonly_fields = (
        "id",
        "type",
        "source",
        "start_date",
        "end_date",
        "published_date",
        "last_modified",
        "sushi_service",
        "subject",
        "message",
    )
    list_filter = ("source", "type")

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def get_actions(self, *args, **kwargs):
        return []



# Generated by Django 4.2.20 on 2025-06-02 21:21

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("celus_registry", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                ("id", models.UUIDField(primary_key=True, serialize=False)),
                ("source", models.CharField(max_length=16)),
                ("type", models.CharField(max_length=12)),
                ("start_date", models.DateField(blank=True, null=True)),
                ("end_date", models.DateField(blank=True, null=True)),
                ("subject", models.CharField(max_length=128)),
                ("message", models.TextField(blank=True)),
                (
                    "published_date",
                    models.DateTimeField(
                        blank=True,
                        help_text="When the notification was published",
                        null=True,
                    ),
                ),
                ("last_modified", models.DateTimeField()),
                (
                    "sushi_service",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications",
                        to="celus_registry.sushiservice",
                    ),
                ),
            ],
        )
    ]



from django.apps import AppConfig


class CelusRegistryConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_celus_registry"
    verbose_name = "Celus Registry"
    label = "celus_registry"



from datetime import timezone

from rest_framework import serializers

from .models import CounterVersionChoices, Notification, Platform, Report, SushiService


class CounterReleaseField(serializers.Field):
    def to_representation(self, value):
        return value.to_string()

    def to_internal_value(self, data):
        return CounterVersionChoices.from_string(data)


class SushiServiceSerializer(serializers.ModelSerializer):
    counter_release = CounterReleaseField()

    class Meta:
        model = SushiService
        fields = (
            "id",
            "counter_release",
            "url",
            "ip_address_authorization",
            "api_key_required",
            "platform_attr_required",
            "requestor_id_required",
        )


class ShortSushiServiceSerializer(serializers.ModelSerializer):
    class Meta:
        model = SushiService
        fields = ("id",)


class ReportSerializer(serializers.ModelSerializer):
    counter_release = CounterReleaseField()

    class Meta:
        model = Report
        fields = ("counter_release", "report_id")


class PlatformSerializer(serializers.ModelSerializer):
    reports = SushiServiceSerializer(many=True, read_only=True)
    sushi_services = serializers.ListField(child=serializers.URLField(), read_only=True)

    class Meta:
        model = Platform
        fields = (
            "id",
            "name",
            "abbrev",
            "reports",
            "content_provider_name",
            "website",
            "sushi_services",
        )


class NotificationSerializer(serializers.ModelSerializer):
    sushi_service = ShortSushiServiceSerializer()
    last_modified = serializers.DateTimeField(
        format="%Y-%m-%dT%H:%M:%S.%fZ", default_timezone=timezone.utc
    )
    published_date = serializers.DateTimeField(
        format="%Y-%m-%dT%H:%M:%S.%fZ", default_timezone=timezone.utc
    )

    class Meta:
        model = Notification
        fields = (
            "id",
            "published_date",
            "last_modified",
            "source",
            "type",
            "start_date",
            "end_date",
            "sushi_service",
            "subject",
            "message",
        )

    def convert_sushi_service_data(self, data):
        if sushi_service_data := data.pop("sushi_service"):
            data["sushi_service_id"] = sushi_service_data["id"]

        return data

    def to_internal_value(self, data):
        # For now we ignore reports and data_host fields
        data.pop("reports")
        data.pop("data_host")
        return self.convert_sushi_service_data(data)

    def to_representation(self, instance):
        res = super().to_representation(instance)
        return self.convert_sushi_service_data(res)



import logging
import typing
from functools import partial

import celery
import requests
from django.conf import settings
from django.db import models, transaction
from django.dispatch import Signal

from .models import CounterVersionChoices, Notification, Platform, Report, SushiService
from .serializers import (
    NotificationSerializer,
    PlatformSerializer,
    ReportSerializer,
    SushiServiceSerializer,
)

logger = logging.getLogger(__name__)


registry_changed = Signal()


def get_base_url():
    url = getattr(settings, "CELUS_REGISTRY_URL", "https://registry.countermetrics.org")
    return f"{url}/api/v1/"


def get_or_none(model_class: typing.Type[models.Model], *args, **kwargs):
    try:
        return model_class.objects.get(*args, **kwargs)
    except model_class.DoesNotExist:
        return None


class Change(typing.NamedTuple):
    key: str
    old: typing.Optional[str] = None
    new: typing.Optional[str] = None
    category: bool = False


class PlatformColumn(typing.NamedTuple):
    name: str
    uuid: str


def get_val(obj, key):
    if hasattr(obj, key):
        return getattr(obj, key)
    if hasattr(obj, "get"):
        return obj.get(key)
    return None


class ChangeTracker:
    PRETTY_NAMES = {"abbrev": "short name"}

    def __init__(self):
        self.updated = {}
        self.created = {}
        self.deleted = {}

    @classmethod
    def check_changes(cls, table, platform, fields, prefix, old, new):
        for name in fields:
            if name == "id":
                continue
            val_old = get_val(old, name)
            val_new = get_val(new, name)
            if name == "counter_release" and isinstance(val_new, str):
                val_new = CounterVersionChoices.from_string(val_new)
            if val_old != val_new:
                if prefix:
                    table.setdefault(platform, []).append(Change(prefix, category=True))
                    prefix = None
                name = cls.PRETTY_NAMES.get(name, name.replace("_", " "))
                if name == "reports":
                    added = ", ".join(sorted(val_new - val_old))
                    added = "\n\n+ " + added if added and val_old else ""
                    removed = ", ".join(sorted(val_old - val_new))
                    removed = "\n\n- " + removed if removed and val_new else ""
                    table.setdefault(platform, []).append(
                        Change(
                            name,
                            ", ".join(sorted(val_old)),
                            ", ".join(sorted(val_new)) + added + removed,
                        )
                    )
                else:
                    table.setdefault(platform, []).append(
                        Change(name, val_old, val_new)
                    )

    def platform_wrapper(self, old, new, platform):
        table = self.updated
        if old is None:
            table = self.created
        elif new is None:
            table = self.deleted
        return partial(self.check_changes, table, platform)


@celery.shared_task
@transaction.atomic
def update_registry_models():
    client = requests.Session()
    modified = update_platforms_and_sushi_services(client)
    modified = modified or update_notifications(client)
    if modified:
        logging.info("Registry changed => triggering signal")
        registry_changed.send(__name__)


def update_platforms_and_sushi_services(client: requests.Session):
    logging.info("Starting to download data from registry")

    url = f"{get_base_url()}/platform/"
    resp = client.get(url)
    logging.debug(f"{url} ({resp.status_code})")
    if resp.status_code != 200:
        logging.warning("Unable to download platforms. Terminating")
        return

    data = resp.json()

    seen_platform_ids = set()
    seen_sushi_service_ids = set()

    tracker = ChangeTracker()

    for platform_data in data:
        serializer = PlatformSerializer(
            get_or_none(Platform, pk=platform_data["id"]), data=platform_data
        )
        serializer.is_valid(raise_exception=True)
        track = tracker.platform_wrapper(
            serializer.instance,
            platform_data,
            PlatformColumn(platform_data["name"], platform_data["id"]),
        )
        track(
            (name for name, field in serializer.fields.items() if not field.read_only),
            "",
            serializer.instance,  # == get_or_none()
            serializer.initial_data,  # == platform_data
        )
        platform = serializer.save()
        seen_platform_ids.add(platform.id)

        # fill in and update report types
        reports = []
        for report_data in platform_data.get("reports", []):
            rep_serializer = ReportSerializer(
                get_or_none(
                    Report,
                    report_id=report_data["report_id"],
                    counter_release=CounterVersionChoices.from_string(
                        report_data["counter_release"]
                    ),
                ),
                data=report_data,
            )
            rep_serializer.is_valid(raise_exception=True)
            report = rep_serializer.save()
            reports.append(report)

        track(
            ("reports",),
            "",
            {"reports": set([r.report_id for r in platform.reports.all()])},
            {"reports": set([r.report_id for r in reports])},
        )
        platform.reports.set(reports)

        # fill in and update sushi services
        sushi_services = []
        for sushi_service_link in platform_data.get("sushi_services", []):
            resp = client.get(sushi_service_link["url"])
            if resp.status_code != 200:
                logging.warning(
                    "Unable to download sushi service '%s'", sushi_service_link["url"]
                )
                raise RuntimeError("Failed to download sushi service.")

            logging.debug(
                "%s C%s",
                sushi_service_link["url"],
                sushi_service_link["counter_release"],
            )
            service_data = resp.json()
            ser_serializer = SushiServiceSerializer(
                get_or_none(SushiService, pk=service_data["id"]), data=service_data
            )
            ser_serializer.is_valid(raise_exception=True)
            track(
                (
                    name
                    for name, field in ser_serializer.fields.items()
                    if not field.read_only
                ),
                f"SUSHI - {ser_serializer.initial_data['url']}",
                ser_serializer.instance,  # == get_or_none()
                ser_serializer.initial_data,  # == platform_data
            )
            sushi_service = ser_serializer.save()
            seen_sushi_service_ids.add(sushi_service.id)
            sushi_services.append(sushi_service)
        platform.sushi_services.set(sushi_services)

    # Clean removed
    removed = Platform.objects.exclude(id__in=seen_platform_ids)
    for platform in removed:
        track = tracker.platform_wrapper(
            platform, None, PlatformColumn(platform.name, platform.id)
        )
        track(
            (
                name
                for name, field in PlatformSerializer().fields.items()
                if not field.read_only
            ),
            "",
            platform,
            None,
        )
        track(
            ("reports",),
            "",
            {"reports": set([r.report_id for r in platform.reports.all()])},
            {"reports": set()},
        )
        for sushi in platform.sushi_services.all():
            track(
                (
                    name
                    for name, field in SushiServiceSerializer().fields.items()
                    if not field.read_only
                ),
                f"SUSHI - {sushi.url}",
                sushi,
                None,
            )
    removed.delete()
    SushiService.objects.exclude(id__in=seen_sushi_service_ids).delete()

    if tracker.updated or tracker.deleted or tracker.created:
        registry_changed.send(
            __name__,
            updated=tracker.updated,
            deleted=tracker.deleted,
            created=tracker.created,
        )


def update_notifications(client: requests.Session) -> bool:
    """returns True if notifications were modified"""
    url = f"{get_base_url()}/notification/"
    modified = False
    seen_ids = set()
    while url:
        resp = client.get(url)
        logging.debug(f"{url} ({resp.status_code})")
        if resp.status_code != 200:
            logging.warning("Unable to download notifications. Terminating")
            return
        data = resp.json()
        for notification_data in data.get("results", []):
            seen_ids.add(notification_data["id"])
            if old_object := Notification.objects.filter(
                id=notification_data["id"]
            ).last():
                serializer = NotificationSerializer(old_object, data=notification_data)
                serializer.is_valid(raise_exception=True)
                old_serializer = NotificationSerializer(old_object)
                record_modified = (
                    modified or old_serializer.data != serializer.validated_data
                )
                if record_modified:
                    serializer.save()
            else:
                serializer = NotificationSerializer(data=notification_data)
                serializer.is_valid(raise_exception=True)
                record_modified = True
                # Create new notification
                serializer.save()

            modified = modified or record_modified
            # Update models
            serializer.save()

        url = data.get("next", None)

    # delete removed notifications
    delete_count, _ = Notification.objects.exclude(pk__in=seen_ids).delete()
    return modified or bool(delete_count)



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models


class CounterVersionChoices(models.IntegerChoices):
    C4 = 4
    C5 = 5
    C51 = 51

    def to_string(self):
        if self == CounterVersionChoices.C4:
            return "4"
        elif self == CounterVersionChoices.C5:
            return "5"
        elif self == CounterVersionChoices.C51:
            return "5.1"

    @classmethod
    def from_string(self, data):
        if data == "4":
            return CounterVersionChoices.C4
        elif data == "5":
            return CounterVersionChoices.C5
        elif data == "5.1":
            return CounterVersionChoices.C51
        raise ValidationError("Wrong counter value")


class Report(models.Model):
    counter_release = models.PositiveSmallIntegerField(
        choices=CounterVersionChoices.choices
    )
    report_id = models.CharField(max_length=20)

    class Meta:
        unique_together = ("counter_release", "report_id")

    def __str__(self):
        return f"{self.report_id} (C{self.counter_release})"


class Platform(models.Model):
    id = models.UUIDField(primary_key=True)  # noqa
    name = models.CharField(max_length=400)
    abbrev = models.CharField(max_length=50, blank=True)
    content_provider_name = models.CharField(max_length=400, blank=True, null=True)  # noqa
    website = models.URLField(blank=True, null=True)  # noqa
    reports = models.ManyToManyField(
        Report, through="ReportToPlatform", related_name="platforms"
    )

    def __str__(self):
        if self.abbrev:
            return f"{self.name} ({self.abbrev})"
        else:
            return self.name

    @property
    def registry_url(self):
        base_url = getattr(
            settings, "CELUS_REGISTRY_URL", "https://registry.countermetrics.org"
        )
        return f"{base_url}/platform/{self.pk}"


class SushiService(models.Model):
    id = models.UUIDField(primary_key=True)  # noqa
    counter_release = models.PositiveSmallIntegerField(
        choices=CounterVersionChoices.choices
    )
    url = models.URLField(null=True, blank=True)  # noqa
    platform = models.ForeignKey(
        Platform, on_delete=models.SET_NULL, related_name="sushi_services", null=True
    )
    ip_address_authorization = models.BooleanField(
        null=True, blank=True, help_text="Access restricted based on IP address"
    )
    api_key_required = models.BooleanField(
        null=True, blank=True, help_text="Is api key required"
    )
    platform_attr_required = models.BooleanField(
        null=True, blank=True, help_text="Is platform attr required"
    )
    requestor_id_required = models.BooleanField(
        null=True, blank=True, help_text="Is requestor_id required"
    )

    def __str__(self):
        return f"{self.url or ''} (C{self.counter_release})"


class ReportToPlatform(models.Model):
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    platform = models.ForeignKey(Platform, on_delete=models.CASCADE)

    def __str__(self):
        return type(self).__name__


class Notification(models.Model):
    # TODO linked ReportViews
    # TODO linked UsageDataHost

    id = models.UUIDField(primary_key=True)
    source = models.CharField(max_length=16)
    type = models.CharField(max_length=12)
    sushi_service = models.ForeignKey(
        SushiService,
        on_delete=models.CASCADE,
        related_name="notifications",
        blank=True,
        null=True,
    )
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)
    subject = models.CharField(max_length=128)
    message = models.TextField(blank=True)
    published_date = models.DateTimeField(
        blank=True, null=True, help_text="When the notification was published"
    )
    last_modified = models.DateTimeField()
